from dataclasses import dataclass
from typing import List

import numpy as np
from skimage.util import view_as_windows
from tqdm import tqdm

from glcm_cupy import *
from glcm_cupy.conf import NO_OF_FEATURES


def glcm_py_3d(ar: np.ndarray, bin_from: int, bin_to: int,
               radius: int = 2,
               step: int = 1):
    return GLCMPy(bin_from=bin_from,
                  bin_to=bin_to,
                  radius=radius,
                  step=step).glcm_3d(ar)


def glcm_py_2d(ar: np.ndarray,
               bin_from: int,
               bin_to: int,
               radius: int = 2,
               step: int = 1):
    return GLCMPy(bin_from=bin_from,
                  bin_to=bin_to,
                  radius=radius,
                  step=step).glcm_2d(ar)


def glcm_py_ij(i: np.ndarray,
               j: np.ndarray,
               bin_from: int, bin_to: int):
    return GLCMPy(bin_from=bin_from,
                  bin_to=bin_to).glcm_ij(i, j)


@dataclass
class GLCMPy:
    bin_from: int
    bin_to: int
    radius: int = 2
    step: int = 1

    @property
    def diameter(self) -> int:
        return self.radius * 2 + 1

    def glcm_ij(self,
                i: np.ndarray,
                j: np.ndarray) -> List[float]:
        """ Calculate the expected GLCM features using Python

        Notes:
            This is to assert the actual values in the tests.
            Technically can be used in production though very slow.

        Args:
            i: Window I, may be ndim
            j: Window J, may be ndim

        Returns:
            List of these features [
                homogeneity, contrast, asm, mean, mean_j,
                var_i, var_j, correlation
            ].
            Values as float.
        """

        i_flat = i.flatten()
        j_flat = j.flatten()
        assert len(i_flat) == len(j_flat), \
            f"The shapes for i {i.shape} != j {j.shape}."

        glcm = np.zeros((self.bin_to, self.bin_to), dtype=float)

        # Populate the GLCM
        for i_, j_ in zip(i_flat, j_flat):
            glcm[i_, j_] += 1
            glcm[j_, i_] += 1

        # Convert to probability
        glcm /= len(i_flat) * 2

        homogeneity = contrast = asm = mean = var = correlation = 0
        for i in range(glcm.shape[0]):
            for j in range(glcm.shape[1]):
                homogeneity += glcm[i, j] / (1 + (i - j) ** 2)
                contrast += glcm[i, j] * (i - j) ** 2
                asm += glcm[i, j] ** 2
                mean += glcm[i, j] * i

        for i in range(glcm.shape[0]):
            for j in range(glcm.shape[1]):
                var += glcm[i, j] * (i - mean) ** 2

        # Variances cannot be 0 else ZeroDivisionError
        if var != 0:
            for i in range(glcm.shape[0]):
                for j in range(glcm.shape[1]):
                    correlation += glcm[i, j] * (i - mean) * (j - mean) / var

        return [homogeneity, contrast, asm, mean, var, correlation]

    def glcm_2d(self, ar: np.ndarray):
        ar = (ar / self.bin_from * self.bin_to).astype(np.uint8)
        ar_w = view_as_windows(ar, (self.diameter, self.diameter))

        def flat(ar: np.ndarray):
            ar = ar.reshape((-1, self.diameter, self.diameter))
            return ar.reshape((ar.shape[0], -1))

        ar_w_i = flat(ar_w[self.step:-self.step, self.step:-self.step])
        ar_w_j_sw = flat(ar_w[self.step * 2:, :-self.step * 2])
        ar_w_j_s = flat(ar_w[self.step * 2:, self.step:-self.step])
        ar_w_j_se = flat(ar_w[self.step * 2:, self.step * 2:])
        ar_w_j_e = flat(ar_w[self.step:-self.step, self.step * 2:])

        feature_ar = np.zeros((ar_w_i.shape[0], 4, NO_OF_FEATURES))

        for j_e, ar_w_j in enumerate(
            (ar_w_j_sw, ar_w_j_s, ar_w_j_se, ar_w_j_e)):
            for e, (i, j) in tqdm(enumerate(zip(ar_w_i, ar_w_j)),
                                  total=ar_w_i.shape[0]):
                feature_ar[e, j_e] = self.glcm_ij(i, j)

        h, w = ar_w.shape[:2]
        feature_ar = feature_ar.mean(axis=1)

        feature_ar = feature_ar.reshape(
            (h - self.step * 2, w - self.step * 2, NO_OF_FEATURES))
        feature_ar[..., CONTRAST] /= (self.bin_to - 1) ** 2
        feature_ar[..., MEAN] /= (self.bin_to - 1)
        feature_ar[..., VAR] /= (self.bin_to - 1) ** 2
        feature_ar[..., CORRELATION] += 1
        feature_ar[..., CORRELATION] /= 2

        return feature_ar

    def glcm_3d(self, ar: np.ndarray):
        return np.stack([self.glcm_2d(ch) for ch in ar])



import cupy as cp

glcm_module = cp.RawModule(
    code=r"""
   #define HOMOGENEITY 0
#define CONTRAST 1
#define ASM 2
#define MEAN 3
#define VAR 4
#define CORRELATION 5
#define NO_OF_FEATURES 6

extern "C" {
    __device__ static inline char atomicAdd(
        unsigned char* address,
        unsigned char val
        )
    {
        // https://stackoverflow.com/questions/5447570/cuda-atomic-operations-on-unsigned-chars
        size_t long_address_modulo = (size_t) address & 3;
        unsigned int* base_address = (unsigned int*) ((char*) address - long_address_modulo);
        unsigned int long_val = (unsigned int) val << (8 * long_address_modulo);
        unsigned int long_old = atomicAdd(base_address, long_val);

        if (long_address_modulo == 3) {
            // the first 8 bits of long_val represent the char value,
            // hence the first 8 bits of long_old represent its previous value.
            return (char) (long_old >> 24);
        } else {
            // bits that represent the char value within long_val
            unsigned int mask = 0x000000ff << (8 * long_address_modulo);
            unsigned int masked_old = long_old & mask;
            // isolate the bits that represent the char value within long_old, add the long_val to that,
            // then re-isolate by excluding bits that represent the char value
            unsigned int overflow = (masked_old + long_val) & ~mask;
            if (overflow) {
                atomicSub(base_address, overflow);
            }
            return (char) (masked_old >> 8 * long_address_modulo);
        }
    }
    __global__ void glcmCreateKernel(
        const unsigned char* windows_i,
        const unsigned char* windows_j,
        const int glcmSize,
        const int noOfValues,
        const int noOfWindows,
        unsigned char* g,
        float* features)
    {
        /**
        =====================================
        Definitions
        =====================================

        windows_i, windows_j = 2D Array. Shape: (noOfValues, noOfWindows)

        It should be an array of windows to be used

        Take for example the following input:

            +---------+  +---------+  +---------+  +---------+
            |         |  |         |  |         |  |         |
            |  3 x 3  |  |         |  |         |  |         |
            |         |  |         |  |         |  |         |
            +---------+  +---------+  +---------+  +---------+

            +---------+  +---------+  +---------+  +---------+
            |         |  |         |  |         |  |         |
            |         |  |         |  |         |  |         |
            |         |  |         |  |         |  |         |
            +---------+  +---------+  +---------+  +---------+

            We don't require the y dimensions. Thus, [3 x 3] x [4 x 2] = 9 x 8

            windows_i will thus be 9 x 8 large. Though in CuPy, we simply flatten everything anyways.m
            It's a simpler way to represent how it works.

        glcmSize = The maximum value of the input + 1, the maximum GLCM size.

        noOfValues = Number of values per window. In the above, it's 3 x 3 = 9

        noOfWindows = Number of windows in total. In the above, it's 4 x 2 = 8

        g = Empty initialized GLCM array. Shape of (glcmSize, glcmSize, noOfWindows)

        features = Empty initialized feature array. Shape of (8, noOfWindows)

        **/

        /**
        =====================================
        Thread ID Calculation
        =====================================
        We are not interested in matching the block/thread dim to
        any dims of the input. We just want to ensure that we have
        enough threads in total.
        **/

        int blockId = blockIdx.y * gridDim.x + blockIdx.x;
        int tid = blockId * blockDim.x + threadIdx.x;

        /**
        =====================================
        TID to respective windows
        =====================================
        TID is 1D, we need to partition them to their windows
        Each window area is `noOfValues` == diameter ^ 2
        We have total of `noOfWindows` number of windows
        Thus, we simply take `wid = tid / noOfValues`
        WID: Window ID

        Note integer division
        Example:
            1  / 50 = 0
            49 / 50 = 0
            50 / 50 = 1
            51 / 50 = 1

        If we have `noOfWindows == 4`, we reject any results >= 4
        Example:
            1   / 50 = 0 (Accept)
            199 / 50 = 3 (Accept)
            200 / 50 = 4 (Reject)
            201 / 50 = 4 (Reject)

        **/

        const int glcmArea = glcmSize * glcmSize;

        int wid_image = tid / noOfValues;
        const static unsigned char x = 1;
        if (tid < noOfValues * noOfWindows)
        {
            unsigned char row = windows_i[tid];
            unsigned char col = windows_j[tid];
            // Remember that the shape of GLCM is (glcmSize, glcmSize, noOfWindows)
            atomicAdd(&(
                g[
                col +
                row * glcmSize +
                wid_image * glcmArea
                ]), x);
            atomicAdd(&(
                g[
                row +
                col * glcmSize +
                wid_image * glcmArea
                ]), x);
        }
    }

    __global__ void glcmFeatureKernel0(
        const unsigned char* g,
        const int glcmSize,
        const int noOfValues,
        const int noOfWindows,
        float* features)
    {

        /**
        ===================================
        TID to respective GLCM i, j, window
        ===================================

        To calculate the window for the tid, we simply divide by the max i & j

        +---------+ +---------+ +---------+
        |         | |         | |         |
        |  i x j  | |         | |         |
        |         | |         | |         |
        +---------+ +---------+ +---------+
        Window 0    Window 1    Window 2    ...

        The area for each GLCM window is glcmSize * glcmSize

        wid = tid / glcmArea
        i = (tid % glcmArea) / glcmSize
        j = (tid % glcmArea) % glcmSize

        **/

        int blockId = blockIdx.y * gridDim.x + blockIdx.x;
        int tid = blockId * blockDim.x + threadIdx.x;

        const int glcmArea = glcmSize * glcmSize;
        const int wid = tid / glcmArea;
        if (wid >= noOfWindows) return;

        const float i = (float)((tid % glcmArea) / glcmSize);
        const float j = (float)((tid % glcmArea) % glcmSize);

        float p = (float)(g[tid]) / (noOfValues * 2);

        /**
        =====================================
        Feature Calculation
        =====================================

        For each feature, we require a wid * NO_OF_FEATURES offset.

        8 x 1 for each GLCM
        +----------------+ +----------------+ +----------------+
        | HOMOGENEITY    | | HOMOGENEITY    | | HOMOGENEITY    |
        | CONTRAST       | | CONTRAST       | | CONTRAST       |
        | ASM            | | ASM            | | ASM            |
        | MEAN           | | MEAN           | | MEAN           |
        | VAR            | | VAR            | | VAR            |
        | CORRELATION    | | CORRELATION    | | CORRELATION    |
        +----------------+ +----------------+ +----------------+
        Window 0           Window 1           Window 2           ...
        **/

        __syncthreads();

        atomicAdd(
            &features[HOMOGENEITY + wid * NO_OF_FEATURES],
            p / (1 + powf((i - j), 2.0f))
        );

        atomicAdd(
            &features[CONTRAST + wid * NO_OF_FEATURES],
            p * powf(i - j, 2.0f)
        );

        atomicAdd(
            &features[ASM + wid * NO_OF_FEATURES],
            powf(p, 2.0f)
        );

        atomicAdd(
            &features[MEAN + wid * NO_OF_FEATURES],
            p * i
        );

    }
    __global__ void glcmFeatureKernel1(
        const unsigned char* g,
        const int glcmSize,
        const int noOfValues,
        const int noOfWindows,
        float* features)
    {
        /**
        =====================================
        Feature Calculation
        =====================================

        See above.
        **/

        int blockId = blockIdx.y * gridDim.x + blockIdx.x;
        int tid = blockId * blockDim.x + threadIdx.x;

        const int glcmArea = glcmSize * glcmSize;
        const int wid = tid / glcmArea;
        if (wid >= noOfWindows) return;

        const float i = (float)((tid % glcmArea) / glcmSize);
        const float j = (float)((tid % glcmArea) % glcmSize);

        float p = (float)(g[tid]) / (noOfValues * 2);

        atomicAdd(
            &features[VAR + wid * NO_OF_FEATURES],
            p * powf((i - features[MEAN + wid * NO_OF_FEATURES]), 2.0f)
        );

    }

    __global__ void glcmFeatureKernel2(
        const unsigned char* g,
        const int glcmSize,
        const int noOfValues,
        const int noOfWindows,
        float* features)
    {
        /**
        =====================================
        Feature Calculation
        =====================================

        See above.
        **/

        int blockId = blockIdx.y * gridDim.x + blockIdx.x;
        int tid = blockId * blockDim.x + threadIdx.x;

        const int glcmArea = glcmSize * glcmSize;
        const int wid = tid / glcmArea;
        if (wid >= noOfWindows) return;

        // As we invert Variance, they should never be 0.
        if (features[VAR + wid * NO_OF_FEATURES] == 0) return;

        const float i = (float)((tid % glcmArea) / glcmSize);
        const float j = (float)((tid % glcmArea) % glcmSize);

        float p = (float)(g[tid]) / (noOfValues * 2);

        atomicAdd(
            &features[CORRELATION + wid * NO_OF_FEATURES],
            p * (i - features[MEAN + wid * NO_OF_FEATURES])
              * (j - features[MEAN + wid * NO_OF_FEATURES])
              / features[VAR + wid * NO_OF_FEATURES]
        );
    }
}
"""
)



from .conf import HOMOGENEITY, CONTRAST, ASM, MEAN, VAR, CORRELATION
from .glcm import GLCM, glcm
from .windowing import Direction, im_shape_after_glcm

__all__ = [
    "glcm",
    "GLCM",
    "HOMOGENEITY",
    "CONTRAST",
    "ASM",
    "MEAN",
    "VAR",
    "CORRELATION",
    "Direction",
    "im_shape_after_glcm"
]



from pathlib import Path

ROOT_DIR = Path(__file__).parent.parent.absolute()

MAX_VALUE_SUPPORTED = 256
NO_OF_VALUES_SUPPORTED = 256 ** 2

MAX_THREADS = 512  # Lowest Maximum supported threads.

NO_OF_FEATURES = 6

# For a 10000 x 256 x 256 GLCM, you need ~ 600MB of memory.
MAX_PARTITION_SIZE = 10000

HOMOGENEITY = 0
CONTRAST = 1
ASM = 2
MEAN = 3
VAR = 4
CORRELATION = 5



from __future__ import annotations

from enum import Enum
from typing import Tuple, List, Iterable

import cupy as cp
import numpy as np
from skimage.util import view_as_windows


class Direction(Enum):
    EAST = 0
    SOUTH_EAST = 1
    SOUTH = 2
    SOUTH_WEST = 3


def make_windows(im: np.ndarray,
                 radius: int,
                 step_size: int,
                 directions: Iterable[Direction] =
                 (Direction.EAST,
                  Direction.SOUTH_EAST,
                  Direction.SOUTH,
                  Direction.SOUTH_WEST)
                 ) -> List[Tuple[np.ndarray, np.ndarray]]:
    """ From a 2D image np.ndarray, convert it into GLCM IJ windows.

    Examples:

        Input 4 x 4 image. Radius = 1. Step Size = 1.

        +-+-+-+-+         +-----+
        |       |         | +---+-+
        |       |  ---->  | |   | |
        |       |         | |   | |
        |       |         +-+---+ |
        +-+-+-+-+           +-----+      flat
          4 x 4           1 x 1 x 3 x 3  ----> 1 x 9
                          +---+   +---+
                          flat    flat

        The output will be flattened on the x,y,

        Input 5 x 5 image. Radius = 1. Step Size = 1.

        1-2-+-+-+-+         1-----+    2-----+
        3 4       |         | +---+-+  | +---+-+  3-----+    4-----+
        |         |  ---->  | |   | |  | |   | |  | +---+-+  | +---+-+
        |         |         | |   | |  | |   | |  | |   | |  | |   | |
        |         |         +-+---+ |  +-+---+ |  | |   | |  | |   | |
        |         |           +-----+    +-----+  +-+---+ |  +-+---+ |
        +-+-+-+-+-+                                 +-----+    +-----+
          4 x 4                         2 x 2 x 3 x 3 ----> 4 x 9
                                        +---+   +---+ flat
                                        flat    flat
    Args:
        im: Input Image
        radius: Radius of Window
        step_size: Step Size between ij pairs
        directions: Directions to pair the windows.

    Returns:
        A List of I, J windows based on the directions.

        For each window:
            The first dimension: xy flat window indexes,
            the last dimension: xy flat indexes within each window.

    """

    # This will yield a shape (window_i, window_j, row, col)
    # E.g. 100x100 with 5x5 window -> 96, 96, 5, 5
    shape = im_shape_after_glcm(im.shape, step_size=step_size, radius=radius)
    if shape[0] <= 0 or shape[1] <= 0:
        raise ValueError(
            f"Step Size & Diameter exceeds size for windowing. "
            f"im.shape[0] {im.shape[0]} "
            f"- 2 * step_size {step_size} "
            f"- 2 * radius {radius} <= 0 or"
            f"im.shape[1] {im.shape[1]} "
            f"- 2 * step_size {step_size} "
            f"- 2 * radius {radius} + 1 <= 0 was not satisfied."
        )

    diameter = radius * 2 + 1
    ij = cp.asarray(view_as_windows(im, (diameter, diameter)))

    ijs: List[Tuple[np.ndarray, np.ndarray]] = []

    for direction in directions:
        i, j = pair_windows(ij, step_size=step_size, direction=direction)

        i = i.reshape((-1, i.shape[-2], i.shape[-1])) \
            .reshape((i.shape[0] * i.shape[1], -1))
        j = j.reshape((-1, j.shape[-2], j.shape[-1])) \
            .reshape((j.shape[0] * j.shape[1], -1))
        ijs.append((i, j))

    return ijs


def pair_windows(ij: np.ndarray,
                 step_size: int,
                 direction: Direction):
    """ Pairs the ij windows

    Notes:

        For an image:

        +-----------+
        |           |
        |  +-----+  |
        |  |     |  |
        |  |     |  |
        |  +-----+  |
        |           |
        +-----------+
                  <-> Distance = Step Size

        The i window will always be the one in the middle
        We move j window around the outer skirts to define the pair.

    Args:
        ij: The ij output from make_windows
        step_size: Step Size between pairs
        direction: Direction to pair

    Returns:

    """
    i = ij[step_size:-step_size, step_size:-step_size]
    if direction == Direction.EAST:
        j = ij[step_size:-step_size, step_size * 2:]
    elif direction == Direction.SOUTH:
        j = ij[step_size * 2:, step_size:-step_size]
    elif direction == Direction.SOUTH_WEST:
        j = ij[step_size * 2:, :-step_size * 2]
    elif direction == Direction.SOUTH_EAST:
        j = ij[step_size * 2:, step_size * 2:]
    else:
        raise ValueError("direction must be of class Direction.")

    return i, j


def im_shape_after_glcm(im_shape: Tuple,
                        step_size: int,
                        radius: int):
    """ Calculate the image shape after GLCM

    Args:
        im_shape: Image Shape before GLCM
        step_size: Step Size
        radius: Radius of GLCM

    Returns:
        Shape of Image after GLCM
    """

    if len(im_shape) < 2:
        raise ValueError(f"Image must be at least 2 dims.")

    return (im_shape[0] - 2 * step_size - 2 * radius,
            im_shape[1] - 2 * step_size - 2 * radius,
            *im_shape[2:])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['glcm_cupy']

package_data = \
{'': ['*']}

install_requires = \
['scikit-image>=0.18.3,<0.19.0', 'tqdm>=4.63.0,<4.64.0']

setup_kwargs = {
    'name': 'glcm-cupy',
    'version': '0.1.6',
    'description': 'Binned GLCM 5 Features implemented in CuPy',
    'long_description': '# GLCM Bin 5 on CuPy\n\nThis is a CuPy reimplementation of my `glcmbin5` [**on my other repository**](https://github.com/Eve-ning/glcmbin5).\n\nThis directly utilizes CUDA to speed up the processing of GLCM.\n\n# Installation\n\n**Python >= 3.7**\n\nFirst, you need to install this\n\n```shell\npip install glcm-cupy\n```\n\nThen, you need **CuPy**.\nYou need to [install CuPy manually](https://docs.cupy.dev/en/stable/install.html), \nas it\'s dependent on the version of CUDA you have.\n\nI recommend using `conda-forge` as it worked for me :)\n\nFor CUDA `11.6`, we use\n```shell\nconda install -c conda-forge cupy cudatoolkit=11.6\n```\n\nReplace the version you have on the arg.\n\n```shell\nconda install -c conda-forge cupy cudatoolkit=__._\n```\n\n# Usage\n\nThe usage is simple:\n\n```pycon\n>>> from glcm_cupy import GLCM\n>>> import numpy as np\n>>> from PIL import Image\n>>> ar = np.asarray(Image.open("image.jpg"))\n>>> ar.shape\n(1080, 1920, 3)\n>>> g = GLCM(...).run(ar)\n>>> g.shape\n(1074, 1914, 3, 8)\n```\n\nThe last dimension of `g` is the GLCM Features.\n\nTo retrieve a specific GLCM Feature:\n\n```pycon\n>>> from glcm_cupy import CONTRAST\n>>> g[..., CONTRAST].shape\n(1074, 1914, 3)\n```\n\nYou may also consider simply `glcm` if you\'re not reusing `GLCM()`\n```pycon\n>>> from glcm_cupy import glcm\n>>> g = glcm(ar, ...)\n```\n\n## **[Example: Processing an Image](examples/process_an_image/main.py)**\n\n## Features\n\nThese are the features implemented.\n\n- `HOMOGENEITY = 0`\n- `CONTRAST = 1`\n- `ASM = 2`\n- `MEAN_I = 3`\n- `MEAN_J = 4`\n- `VAR_I = 5`\n- `VAR_J = 6`\n- `CORRELATION = 7`\n\nDon\'t see one you need? Raise an issue, I\'ll (hopefully) add it.\n\n## Radius & Step Size\n\n- The radius defines the window radius for each GLCM window.\n- The step size defines the distance between each window.\n  - If it\'s diagonal, it treats a diagonal step as 1. It\'s not the euclidean distance.\n\n## Binning\n\nTo reduce GLCM processing time, you can specify `bin_from` & `bin_to`.\n\nThis will bin the image from a range to another.\n\nI highly recommend using this to reduce time taken before raising it.\n\nE.g.\n\n> I have an RGB image with a max value of 255.\n> \n> I limit the max value to 31. This reduces the processing time.\n> \n> `GLCM(..., bin_from=256, bin_to=32).run(ar)`\n\nThe lower the max value, the smaller the GLCM required. Thus allowing for\nmore GLCMs to run concurrently.\n\n## Direction\n\nBy default we have the following directions to run GLCM on.\n\n- East: `Direction.EAST`\n- South East: `Direction.SOUTH_EAST`\n- South: `Direction.SOUTH`\n- South West: `Direction.SOUTH_WEST`\n\nFor each direction, the GLCM will be bi-directional.\n\nWe can specify only certain directions here.\n\n```pycon\n>>> from glcm_cupy import GLCM\n>>> GLCM()\n>>> g = GLCM(directions=(Direction.SOUTH_WEST, Direction.SOUTH))\n```\n\nThe result of these directions will be averaged together.\n\n# Notes\n\n> Q: Why did my image shrink?\n> \n> The image shrunk due to `step_size` & `radius`.\n> \n> The amount of shrink per XY Dimension is\n> `size - 2 * step_size - 2 * radius`\n\n> Q: What\'s the difference between this and `glcmbin5`?\n> \n> This is the faster one, and easier to use.\n> I highly recommend avoiding `glcmbin5` as it has C++, which means you need to compile manually.\n> \n> It\'s the first version of GLCM I made.\n\n## CUDA Notes\n\n### Why is the kernel split into 4?\n\nThe kernel is split into 4 sections\n\n1) GLCM Creation\n2) Features (ASM, Contrast, Homogeneity, GLCM Mean I, GLCM Mean J)\n3) Features (GLCM Variance I, GLCM Variance J)\n4) Features (GLCM Correlation)\n\nThe reason why it\'s split is due to (2) being reliant on (1), and (3) on (2), ... .\n\nThere are some other solutions tried\n\n1) `__syncthreads()` will not work as we require to sync all blocks.\n    1) We can\'t put all calculations in a block due to the thread limit of 512, 1024, 2048.\n    2) We require 256 * 256 threads minimum to support a GLCM of max value 255.\n2) **Cooperative Groups** imposes a 24 block limit.\n\nThus, the best solution is to split the kernel.\n\n### Atomic Add\n\nThreads cannot write to a single pointer in parallel, information will be overwritten and lost. This is the **Race\nCondition**.\n\nIn order to avoid this, we use [**Atomic\nFunctions**](https://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#atomic-functions).\n\n> ... it is guaranteed to be performed without interference from other threads\n\n### Custom Atomic Add\n\nCurrently `atomicAdd()` doesn\'t have the signature to support `uint8` or `unsigned char`. We get this implementation\nfrom this [**StackOverflow\nAnswer**](https://stackoverflow.com/questions/5447570/cuda-atomic-operations-on-unsigned-chars)\n\n',
    'author': 'Evening',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Eve-ning/glcmbin5_cupy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



from __future__ import annotations

import math
from typing import Tuple, List, Iterable

import cupy as cp
import numpy as np
from tqdm import tqdm

from glcm_cupy.conf import *
from glcm_cupy.kernel import glcm_module
from glcm_cupy.windowing import make_windows, im_shape_after_glcm, Direction


def glcm(
    im: np.ndarray,
    step_size: int = 1,
    radius: int = 2,
    bin_from: int = 256,
    bin_to: int = 256,
    directions: Iterable[Direction] = (Direction.EAST,
                                       Direction.SOUTH_EAST,
                                       Direction.SOUTH,
                                       Direction.SOUTH_WEST),
    max_partition_size: int = MAX_PARTITION_SIZE,
    normalize_features: bool = True) -> Tuple[List[str] ,np.ndarray]:
    """
    Examples:
        To scale down the image from a 128 max value to 32, we use
        bin_from = 128, bin_to = 32.

        The range will collapse from 128 to 32.

        This thus optimizes the GLCM speed.

    Args:
        im: Image to Process
        step_size: Stride Between GLCMs
        radius: Radius of Window
        bin_from: Binarize from.
        bin_to: Binarize to.
        directions: Directions to pair the windows.
        max_partition_size: Maximum number of windows to parse at once
        normalize_features: Whether to normalize features to [0, 1]

    Returns:


    """
    return GLCM(step_size, radius, bin_from, bin_to,
                directions, max_partition_size, normalize_features).run(im)


class GLCM:
    def __init__(self,
                 step_size: int = 1,
                 radius: int = 2,
                 bin_from: int = 256,
                 bin_to: int = 256,
                 directions: Iterable[Direction] = (Direction.EAST,
                                                    Direction.SOUTH_EAST,
                                                    Direction.SOUTH,
                                                    Direction.SOUTH_WEST),
                 max_partition_size: int = MAX_PARTITION_SIZE,
                 max_threads: int = MAX_THREADS,
                 normalize_features: bool = True
                 ):
        """ Initialize Settings for GLCM

        Examples:
            To scale down the image from a 128 max value to 32, we use
            bin_from = 128, bin_to = 32.

            The range will collapse from 128 to 32.

            This thus optimizes the GLCM speed.

        Args:
            step_size: Stride Between GLCMs
            radius: Radius of Window
            bin_from: Binarize from.
            bin_to: Binarize to.
            directions: Directions to pair the windows.
            max_partition_size: Maximum number of windows to parse at once
            max_threads: Maximum number of threads to use per block
            normalize_features: Whether to normalize features to [0, 1]
        """
        self.step_size = step_size
        self.radius = radius
        self.bin_from = bin_from
        self.bin_to = bin_to
        self.directions = directions
        self.max_partition_size = max_partition_size
        self.max_threads = max_threads
        self.normalize_features = normalize_features
        self.progress = None

        self.glcm = cp.zeros(
            (self.max_partition_size, self.bin_to, self.bin_to),
            dtype=cp.uint8)
        self.features = cp.zeros((self.max_partition_size, NO_OF_FEATURES),
                                 dtype=cp.float32)
        if self.radius < 0:
            raise ValueError(
                f"Radius {radius} should be > 0)"
            )
        if bin_to not in range(2, MAX_VALUE_SUPPORTED + 1):
            raise ValueError(
                f"Target Bins {bin_to} should be "
                f"[2, {MAX_VALUE_SUPPORTED}]. "
            )
        if self.step_size <= 0:
            raise ValueError(
                f"Step Size {step_size} should be >= 1"
            )

        self.glcm_create_kernel = \
            glcm_module.get_function('glcmCreateKernel')
        self.glcm_feature_kernel_0 = \
            glcm_module.get_function('glcmFeatureKernel0')
        self.glcm_feature_kernel_1 = \
            glcm_module.get_function('glcmFeatureKernel1')
        self.glcm_feature_kernel_2 = \
            glcm_module.get_function('glcmFeatureKernel2')

    @property
    def _diameter(self):
        return self.radius * 2 + 1

    def run(self, im: np.ndarray):
        """ Executes running GLCM. Returns the GLCM Feature array

        Args:
            im: Image to process. Can be 2D or 3D.

        Returns:
            An np.ndarray of Shape,
             3D: (rows, cols, channels, features),
             2D: (rows, cols, features)

        """
        shape = im_shape_after_glcm(im.shape, self.step_size, self.radius)
        glcm_cells = np.prod(shape) * len(self.directions)
        self.progress = tqdm(total=glcm_cells, desc="GLCM Progress",
                             unit=" Cells",
                             unit_scale=True)

        if im.ndim == 2:
            return self._from_2dimage(im)
        elif im.ndim == 3:
            return self._from_3dimage(im)
        else:
            raise ValueError("Only 2D or 3D images allowed.")

    def _from_3dimage(self, im: np.ndarray) -> np.ndarray:
        """ Generates the GLCM from a multi band image

        Args:
            im: A 3 dim image as an ndarray

        Returns:
            The GLCM Array 4dim with shape
                rows, cols, channel, feature
        """

        glcm_chs = []
        for ch in range(im.shape[-1]):
            glcm_chs.append(self._from_2dimage(im[..., ch]))

        return np.stack(glcm_chs, axis=2)

    def _from_2dimage(self,
                      im: np.ndarray) -> np.ndarray:
        """ Generates the GLCM from a single band image

        Notes:
            Partitions the image into blocks of PARTITION_SIZE,
            reducing allocated GLCM Memory.

        Args:
            im: Image in np.ndarray.

        Returns:
            The GLCM Array 3dim with shape
                rows, cols, feature
        """

        if im.ndim != 2:
            raise ValueError(f"Image must be 2 dimensional. "
                             f"im.ndim={im.ndim}")

        im = self._binner(im, self.bin_from, self.bin_to)

        # Both dims are xy flattened.
        # windows.shape == [window_ix, cell_ix]
        dirs_ij = make_windows(im,
                               self.radius,
                               self.step_size,
                               self.directions)
        dirs_ij: List[Tuple[np.ndarray, np.ndarray]]
        glcm_features_dirs = []

        # For each direction
        for dir, dir_ij in zip(self.directions, dirs_ij):

            windows_i, windows_j = dir_ij
            windows_count = windows_i.shape[0]
            glcm_features_dir = cp.zeros(
                (windows_count, NO_OF_FEATURES),
                dtype=cp.float32
            )

            # If 45000 windows, 10000 partition size,
            # We have 5 partitions (10K, 10K, 10K, 10K, 5K)
            partition_count = math.ceil(
                windows_count / self.max_partition_size)

            for partition in range(partition_count):
                with cp.cuda.Device() as dev:
                    # As above, we have an uneven partition
                    # Though [1,2,3][:5] == [1,2,3]
                    # This means windows_part_i will be correctly partitioned.
                    start = partition * self.max_partition_size
                    end = (partition + 1) * self.max_partition_size
                    part_i = windows_i[start:end]
                    part_j = windows_j[start:end]

                    # We don't need to figure out the leftover partition size
                    # We can simply yield the length of the leftover partition
                    windows_part_count = part_i.shape[0]

                    self.glcm[:] = 0
                    glcm_features_dir[start:start + windows_part_count] = \
                        self.run_ij(
                            part_i,
                            part_j
                        )
                    dev.synchronize()
                    self.progress.update(windows_part_count)

            shape = im_shape_after_glcm(im.shape,
                                        step_size=self.step_size,
                                        radius=self.radius)
            glcm_features_dirs.append(
                glcm_features_dir
                    .reshape(shape[0], shape[1], NO_OF_FEATURES)
                    .get()
            )

        ar = np.stack(glcm_features_dirs).mean(axis=0)

        if self.normalize_features:
            # This scales the glcm features to [0, 1]
            ar[..., CONTRAST] /= (self.bin_to - 1) ** 2
            ar[..., MEAN] /= (self.bin_to - 1)
            ar[..., VAR] /= (self.bin_to - 1) ** 2
            ar[..., CORRELATION] += 1
            ar[..., CORRELATION] /= 2
        return ar

    def run_ij(self,
               i: np.ndarray,
               j: np.ndarray):
        """ Generate the GLCM from the I J Window

        Examples:

            >>> ar_0 = np.random.randint(0, 100, 10, dtype=np.uint8)
            >>> ar_1 = np.random.randint(0, 100, 10, dtype=np.uint8)
            >>> g = GLCM().run_ij(ar_0[...,np.newaxis], ar_1[...,np.newaxis])

        Notes:
            i must be the same shape as j
            i.shape, j.shape should be [partition_size, window_size]

            For example, if you want to process 100 windows at once,
            each window with 25 cells, you should have
            i.shape == j.shape == (100, 25)

        Args:
            i: 1st np.ndarray
            j: 2nd np.ndarray

        Returns:
            The GLCM feature array, of size [partition_size, 8]

        """
        if i.ndim != 2 or j.ndim != 2:
            raise ValueError(
                f"The input dimensions must be 2. "
                f"i.ndim=={i.ndim}, j.ndim=={j.ndim}. "
                "The 1st Dim is the partitioned windows flattened, "
                "The 2nd is the window cells flattened"
            )

        if i.shape != j.shape:
            raise ValueError(f"Shape of i {i.shape} != j {j.shape}")

        # Reset data
        self.glcm[:] = 0
        self.features[:] = 0

        no_of_windows = i.shape[0]
        no_of_values = self._diameter ** 2

        if i.dtype != np.uint8 or j.dtype != np.uint8:
            raise ValueError(
                f"Image dtype must be np.uint8,"
                f" i: {i.dtype} j: {j.dtype}"
            )

        grid = self._calc_grid_size(no_of_windows,
                                    self.bin_to,
                                    self.max_threads)
        self.glcm_create_kernel(
            grid=grid,
            block=(self.max_threads,),
            args=(
                i,
                j,
                self.bin_to,
                no_of_values,
                no_of_windows,
                self.glcm,
                self.features
            )
        )

        feature_args = dict(
            grid=grid, block=(self.max_threads,),
            args=(self.glcm,
                  self.bin_to,
                  no_of_values,
                  no_of_windows,
                  self.features)
        )
        self.glcm_feature_kernel_0(**feature_args)
        self.glcm_feature_kernel_1(**feature_args)
        self.glcm_feature_kernel_2(**feature_args)

        return self.features[:no_of_windows]

    @staticmethod
    def _calc_grid_size(
        window_count: int,
        glcm_size: int,
        thread_per_block: int = MAX_THREADS
    ) -> Tuple[int, int]:
        """ Calculates the required grid size

        Notes:
            There's 2 points where the number of threads

        Returns:
            The optimal minimum grid shape for GLCM

        """

        # TODO: Do we need to div by thread_per_block?
        # Blocks to support features
        blocks_req_glcm_features = \
            window_count * glcm_size * glcm_size / thread_per_block

        # Blocks to support glcm populating
        blocks_req_glcm_populate = glcm_size * window_count

        # Take the maximum
        blocks_req = max(blocks_req_glcm_features, blocks_req_glcm_populate)

        # We split it to 2 dims: A -> B x B
        b = int(blocks_req ** 0.5) + 1
        return b, b

    @staticmethod
    def _binner(im: np.ndarray, bin_from: int, bin_to: int) -> np.ndarray:
        """ Bins an image from a certain bin to another

        Args:
            im: Image as np.ndarray
            bin_from: From the Bin of input image
            bin_to: To the Bin of output image

        Returns:
            Binned Image

        """
        return (im.astype(np.float32) / bin_from * bin_to).astype(np.uint8)
